import numpy as np

from .grouped_array import GroupedArray


class BaseLocalScaler:
    scaler_type: str

    def fit(self, ga: GroupedArray) -> "BaseLocalScaler":
        self.stats_ = ga.scaler_fit(self.scaler_type)
        return self

    def transform(self, ga: GroupedArray) -> np.ndarray:
        return ga.scaler_transform(self.stats_)

    def inverse_transform(self, ga: GroupedArray) -> np.ndarray:
        return ga.scaler_inverse_transform(self.stats_)


class LocalMinMaxScaler(BaseLocalScaler):
    scaler_type = "MinMax"


class LocalStandardScaler(BaseLocalScaler):
    scaler_type = "Standard"


class LocalRobustScaler(BaseLocalScaler):
    def __init__(self, scale: str):
        if scale == "iqr":
            self.scaler_type = "RobustIqr"
        else:
            self.scaler_type = "RobustMad"



import os

import toml


if os.environ["CIBW_BUILD"] == "cp38-macosx_arm64":
    with open("pyproject.toml", "rt") as f:
        cfg = toml.load(f)
    cfg["tool"]["scikit-build"]["cmake"]["args"] = ["-DUSE_OPENMP=OFF"]
    with open("pyproject.toml", "wt") as f:
        toml.dump(cfg, f)



__all__ = [
    "Lag",
    "RollingMean",
    "RollingStd",
    "RollingMin",
    "RollingMax",
    "SeasonalRollingMean",
    "SeasonalRollingStd",
    "SeasonalRollingMin",
    "SeasonalRollingMax",
    "ExpandingMean",
    "ExpandingStd",
    "ExpandingMin",
    "ExpandingMax",
    "ExponentiallyWeightedMean",
]

import abc
from typing import Callable, Optional

import numpy as np

from .grouped_array import GroupedArray


class BaseLagTransform(abc.ABC):
    @abc.abstractmethod
    def transform(self, ga: GroupedArray) -> np.ndarray:
        ...

    @abc.abstractmethod
    def update(self, ga: GroupedArray) -> np.ndarray:
        ...


class Lag(BaseLagTransform):
    def __init__(self, lag: int):
        self.lag = lag

    def transform(self, ga: GroupedArray) -> np.ndarray:
        return ga.lag_transform(self.lag)

    def update(self, ga: GroupedArray) -> np.ndarray:
        return ga.take_from_groups(self.lag - 1)


class RollingBase(BaseLagTransform):
    stat_name: str
    lag: int
    window_size: int
    min_samples: int

    def __init__(self, lag: int, window_size: int, min_samples: Optional[int] = None):
        self.lag = lag
        if min_samples is None:
            min_samples = window_size
        if min_samples > window_size:
            min_samples = window_size
        self.window_size = window_size
        self.min_samples = min_samples

    def transform(self, ga: GroupedArray) -> np.ndarray:
        return ga.rolling_transform(
            self.stat_name, self.lag, self.window_size, self.min_samples
        )

    def update(self, ga: GroupedArray) -> np.ndarray:
        return ga.rolling_update(
            self.stat_name, self.lag - 1, self.window_size, self.min_samples
        )


class RollingMean(RollingBase):
    stat_name = "Mean"


class RollingStd(RollingBase):
    stat_name = "Std"


class RollingMin(RollingBase):
    stat_name = "Min"


class RollingMax(RollingBase):
    stat_name = "Max"


class SeasonalRollingBase(RollingBase):
    season_length: int

    def __init__(
        self,
        lag: int,
        season_length: int,
        window_size: int,
        min_samples: Optional[int] = None,
    ):
        super().__init__(lag, window_size, min_samples)
        self.season_length = season_length

    def transform(self, ga: GroupedArray) -> np.ndarray:
        return ga.seasonal_rolling_transform(
            self.stat_name,
            self.lag,
            self.season_length,
            self.window_size,
            self.min_samples,
        )

    def update(self, ga: GroupedArray) -> np.ndarray:
        return ga.seasonal_rolling_update(
            self.stat_name,
            self.lag - 1,
            self.season_length,
            self.window_size,
            self.min_samples,
        )


class SeasonalRollingMean(SeasonalRollingBase):
    stat_name = "Mean"


class SeasonalRollingStd(SeasonalRollingBase):
    stat_name = "Std"


class SeasonalRollingMin(SeasonalRollingBase):
    stat_name = "Min"


class SeasonalRollingMax(SeasonalRollingBase):
    stat_name = "Max"


class ExpandingBase(BaseLagTransform):
    def __init__(self, lag: int):
        self.lag = lag


class ExpandingMean(ExpandingBase):
    def transform(self, ga: GroupedArray) -> np.ndarray:
        self.n = np.empty_like(ga.data, shape=len(ga))
        out = ga.expanding_transform_with_aggs("Mean", self.lag, self.n)
        self.cumsum = out[ga.indptr[1:] - 1] * self.n
        return out

    def update(self, ga: GroupedArray) -> np.ndarray:
        self.n += 1
        self.cumsum += ga.take_from_groups(self.lag - 1)
        return self.cumsum / self.n


class ExpandingStd(ExpandingBase):
    def transform(self, ga: GroupedArray) -> np.ndarray:
        self.stats = np.empty_like(ga.data, shape=(len(ga), 3))
        out = ga.expanding_transform_with_aggs("Std", self.lag, self.stats)
        return out

    def update(self, ga: GroupedArray) -> np.ndarray:
        x = ga.take_from_groups(self.lag - 1)
        self.stats[:, 0] += 1.0
        n = self.stats[:, 0]
        prev_avg = self.stats[:, 1].copy()
        self.stats[:, 1] = prev_avg + (x - prev_avg) / n
        self.stats[:, 2] += (x - prev_avg) * (x - self.stats[:, 1])
        self.stats[:, 2] = np.maximum(self.stats[:, 2], 0.0)
        return np.sqrt(self.stats[:, 2] / (n - 1))


class ExpandingComp(ExpandingBase):
    stat: str
    comp_fn: Callable

    def transform(self, ga: GroupedArray) -> np.ndarray:
        out = ga.expanding_transform(self.stat, self.lag)
        self.stats = out[ga.indptr[1:] - 1]
        return out

    def update(self, ga: GroupedArray) -> np.ndarray:
        self.stats = self.comp_fn(self.stats, ga.take_from_groups(self.lag - 1))
        return self.stats


class ExpandingMin(ExpandingComp):
    stat = "Min"
    comp_fn = np.minimum


class ExpandingMax(ExpandingComp):
    stat = "Max"
    comp_fn = np.maximum


class ExponentiallyWeightedMean(BaseLagTransform):
    def __init__(self, lag: int, alpha: float):
        self.lag = lag
        self.alpha = alpha

    def transform(self, ga: GroupedArray) -> np.ndarray:
        out = ga.exponentially_weighted_transform("Mean", self.lag, self.alpha)
        self.ewm = out[ga.indptr[1:] - 1]
        return out

    def update(self, ga: GroupedArray) -> np.ndarray:
        x = ga.take_from_groups(self.lag - 1)
        self.ewm = self.alpha * x + (1 - self.alpha) * self.ewm
        return self.ewm



import ctypes
import platform
import sys

import numpy as np

if sys.version_info < (3, 10):
    from importlib_resources import files
else:
    from importlib.resources import files


if platform.system() in ("Windows", "Microsoft"):
    prefix = "Release"
    extension = "dll"
else:
    prefix = ""
    extension = "so"

_LIB = ctypes.CDLL(
    str(files("coreforecast").joinpath("lib", prefix, f"libcoreforecast.{extension}"))
)


def _data_as_void_ptr(arr: np.ndarray):
    return arr.ctypes.data_as(ctypes.POINTER(ctypes.c_void_p))


class GroupedArray:
    def __init__(self, data: np.ndarray, indptr: np.ndarray, num_threads: int = 1):
        data = np.ascontiguousarray(data, dtype=data.dtype)
        if data.dtype == np.float32:
            self.prefix = "GroupedArrayFloat32"
        elif data.dtype == np.float64:
            self.prefix = "GroupedArrayFloat64"
        else:
            self.prefix = "GroupedArrayFloat32"
            data = data.astype(np.float32)
        self.data = data
        if indptr.dtype != np.int32:
            indptr = indptr.astype(np.int32)
        self.indptr = indptr
        self._handle = ctypes.c_void_p()
        _LIB[f"{self.prefix}_Create"](
            _data_as_void_ptr(data),
            ctypes.c_int32(data.size),
            indptr.ctypes.data_as(ctypes.POINTER(ctypes.c_int32)),
            ctypes.c_int(indptr.size),
            ctypes.c_int(num_threads),
            ctypes.byref(self._handle),
        )

    def __del__(self):
        _LIB[f"{self.prefix}_Delete"](self._handle)

    def __len__(self):
        return self.indptr.size - 1

    def __getitem__(self, i):
        return self.data[self.indptr[i] : self.indptr[i + 1]]

    def scaler_fit(self, scaler_type: str) -> np.ndarray:
        stats = np.full_like(self.data, np.nan, shape=(len(self), 2))
        _LIB[f"{self.prefix}_{scaler_type}ScalerStats"](
            self._handle,
            _data_as_void_ptr(stats),
        )
        return stats

    def scaler_transform(self, stats: np.ndarray) -> np.ndarray:
        out = np.full_like(self.data, np.nan)
        _LIB[f"{self.prefix}_ScalerTransform"](
            self._handle,
            _data_as_void_ptr(stats),
            _data_as_void_ptr(out),
        )
        return out

    def scaler_inverse_transform(self, stats: np.ndarray) -> np.ndarray:
        out = np.empty_like(self.data)
        _LIB[f"{self.prefix}_ScalerInverseTransform"](
            self._handle,
            _data_as_void_ptr(stats),
            _data_as_void_ptr(out),
        )
        return out

    def take_from_groups(self, k: int) -> np.ndarray:
        out = np.empty_like(self.data, shape=len(self))
        _LIB[f"{self.prefix}_TakeFromGroups"](
            self._handle,
            ctypes.c_int(k),
            _data_as_void_ptr(out),
        )
        return out

    def lag_transform(self, lag: int) -> np.ndarray:
        out = np.full_like(self.data, np.nan)
        _LIB[f"{self.prefix}_LagTransform"](
            self._handle,
            ctypes.c_int(lag),
            _data_as_void_ptr(out),
        )
        return out

    def rolling_transform(
        self, stat_name: str, lag: int, window_size: int, min_samples: int
    ) -> np.ndarray:
        out = np.full_like(self.data, np.nan)
        _LIB[f"{self.prefix}_Rolling{stat_name}Transform"](
            self._handle,
            ctypes.c_int(lag),
            ctypes.c_int(window_size),
            ctypes.c_int(min_samples),
            _data_as_void_ptr(out),
        )
        return out

    def rolling_update(
        self, stat_name: str, lag: int, window_size: int, min_samples: int
    ) -> np.ndarray:
        out = np.empty_like(self.data, shape=len(self))
        _LIB[f"{self.prefix}_Rolling{stat_name}Update"](
            self._handle,
            ctypes.c_int(lag),
            ctypes.c_int(window_size),
            ctypes.c_int(min_samples),
            _data_as_void_ptr(out),
        )
        return out

    def seasonal_rolling_transform(
        self,
        stat_name: str,
        lag: int,
        season_length: int,
        window_size: int,
        min_samples: int,
    ) -> np.ndarray:
        out = np.full_like(self.data, np.nan)
        _LIB[f"{self.prefix}_SeasonalRolling{stat_name}Transform"](
            self._handle,
            ctypes.c_int(lag),
            ctypes.c_int(season_length),
            ctypes.c_int(window_size),
            ctypes.c_int(min_samples),
            _data_as_void_ptr(out),
        )
        return out

    def seasonal_rolling_update(
        self,
        stat_name: str,
        lag: int,
        season_length: int,
        window_size: int,
        min_samples: int,
    ) -> np.ndarray:
        out = np.empty_like(self.data, shape=len(self))
        _LIB[f"{self.prefix}_SeasonalRolling{stat_name}Update"](
            self._handle,
            ctypes.c_int(lag),
            ctypes.c_int(season_length),
            ctypes.c_int(window_size),
            ctypes.c_int(min_samples),
            _data_as_void_ptr(out),
        )
        return out

    def expanding_transform_with_aggs(
        self,
        stat_name: str,
        lag: int,
        aggs: np.ndarray,
    ) -> np.ndarray:
        out = np.full_like(self.data, np.nan)
        _LIB[f"{self.prefix}_Expanding{stat_name}Transform"](
            self._handle,
            ctypes.c_int(lag),
            _data_as_void_ptr(out),
            _data_as_void_ptr(aggs),
        )
        return out

    def expanding_transform(
        self,
        stat_name: str,
        lag: int,
    ) -> np.ndarray:
        out = np.full_like(self.data, np.nan)
        _LIB[f"{self.prefix}_Expanding{stat_name}Transform"](
            self._handle,
            ctypes.c_int(lag),
            _data_as_void_ptr(out),
        )
        return out

    def exponentially_weighted_transform(
        self,
        stat_name: str,
        lag: int,
        alpha: float,
    ) -> np.ndarray:
        out = np.full_like(self.data, np.nan)
        if self.prefix == "GroupedArrayFloat32":
            alpha = ctypes.c_float(alpha)
        else:
            alpha = ctypes.c_double(alpha)
        _LIB[f"{self.prefix}_ExponentiallyWeighted{stat_name}Transform"](
            self._handle,
            ctypes.c_int(lag),
            alpha,
            _data_as_void_ptr(out),
        )
        return out
